{% extends "admin/submit_line.html" %}
{% block submit-row %}
  {% if include_version_link %}
    <a href="{{ version_link_url }}" class="closelink" title="{{ version_link_title }}">{{ version_link_label }}</a>
  {% endif %}
  {% if include_create_draft_link %}
    <input type="submit" value="{{ create_draft_link_label }}" name="_createdraft">
  {% endif %}
  {% if include_publish_draft_link %}
    <input type="submit" value="{{ publish_draft_link_title }}" name="_publishdraft" {% if publish_draft_link_disabled %}disabled{% endif %}>
  {% endif %}
  {{ block.super }}
{% endblock %}
